import os
from typing import Any

import pandas as pd
from dash import dcc

from config import settings
from resources.dataframe import xlsx_student_data_to_dataframe, group_students_by_day


def calculate_data_table(
    dataframe_dict: dict[str, pd.DataFrame]
) -> list[dict[str, str | int | float]]:
    """Generate list of dicts for data data_table info.

    Args:
        dataframe_dict (dict[str, pd.DataFrame]): Dict of key(day)-value(DataFrame) pairs
        relating to student info for a particular subject.

    Returns:
        student_data (list[dict[str, str | int | float]]): List of dictionaries representing
        statistics for a particular student.

    """
    student_data = []

    for subject, df in dataframe_dict.items():
        student_names = [
            col for col in df.columns if col not in ["date", "lesson time (minutes)", "day"]
        ]
        for name in student_names:
            temp_df = df[["date", "lesson time (minutes)", name]]
            temp_df = temp_df[temp_df[name].notna()]
            missed_lessons = sum(temp_df[name] == 0)
            total_lessons = len(temp_df)
            total_minutes = sum(temp_df["lesson time (minutes)"])
            total_minutes_missed = sum(temp_df["lesson time (minutes)"]) - sum(temp_df[name])

            student_data.append(
                {
                    "name": name,
                    "total_lessons": len(temp_df),
                    "missed_lessons": sum(temp_df[name] == 0),
                    "attended_lessons": total_lessons - missed_lessons,
                    "total_minutes_missed": total_minutes - sum(temp_df[name]),
                    "attendance_percentage": 100
                    - round(100 * (total_minutes_missed / total_minutes), 2),
                    "subject": subject,
                }
            )

    return student_data


def generate_subject_stats_dict() -> tuple[dict[str, dict[str, Any]], int, int, int]:
    """Generate stats objects related to all subjects taught.

    Reads the .xlsx file and generates statistics related to student data.

    Returns
        stats (tuple[dict[str, dict[str, Any]], int, int, int]): dict of subject stats, number of
        days taught, number of students taught, number of subjects taught.
    """
    xl_file = pd.ExcelFile(os.path.join(settings.DATA_DIR, settings.FILE_NAME))

    dfs_dict = {sheet_name: xl_file.parse(sheet_name) for sheet_name in xl_file.sheet_names}

    # dfs = process_dataframes(dataframe_dict=dfs_dict)

    subjects = list(dfs_dict.keys())

    subject_stats: dict[str, Any] = {}

    num_students = 0
    total_student_hours = 0
    total_missed_time = 0
    total_missed_days = 0
    days_taught = set()

    for subject in subjects:
        df = xlsx_student_data_to_dataframe(subject=subject)
        df_by_days = group_students_by_day(df=df)

        subject_stats[subject] = {}

        for day, subject_df in df_by_days.items():
            days_taught.add(day)
            day_student_hours = 0
            day_missed_time = 0
            day_missed_days = 0
            day_num_students = 0

            subject_stats[subject][day] = {}
            student_names = [
                name
                for name in subject_df.columns
                if name not in ["date", "lesson time (minutes)", "day"]
            ]

            day_num_students += len(student_names)
            num_students += len(student_names)

            for name in student_names:
                temp_df = subject_df[["lesson time (minutes)", name]].copy()
                temp_df.dropna(inplace=True)

                # Number of potential teaching minutes student could have attended
                teaching_minutes = round(sum(temp_df["lesson time (minutes)"]), 2)

                # Number of teaching minutes missed by student
                total_minutes = temp_df["lesson time (minutes)"] - temp_df[name]
                missing_minutes = round(sum(total_minutes), 2)

                # Days student did not attend
                missed_days = sum(temp_df[name] == 0)

                attendance_percentage = round(
                    100 * (1 - ((missing_minutes / teaching_minutes))), 2
                )

                subject_stats[subject][day][name] = {
                    "day": day,
                    "teaching_minutes": teaching_minutes,
                    "missed_days": missed_days,
                    "missed_minutes": missing_minutes,
                    "attendance_percentage": attendance_percentage,
                }

                day_missed_time += missing_minutes
                day_student_hours += sum(temp_df["lesson time (minutes)"])
                day_missed_days += missed_days

                total_missed_time += missing_minutes
                total_student_hours += sum(temp_df["lesson time (minutes)"])
                total_missed_days += missed_days

            subject_stats[subject][day]["total_missed_time"] = day_missed_time
            subject_stats[subject][day]["total_student_hours"] = day_student_hours
            subject_stats[subject][day]["total_missed_days"] = day_missed_days
            subject_stats[subject][day]["total_students"] = day_num_students

    subject_stats["total_missed_time"] = total_missed_time
    subject_stats["total_student_hours"] = total_student_hours
    subject_stats["total_missed_days"] = total_missed_days

    return subject_stats, len(days_taught), num_students, len(subjects)


def create_plots(stats_dict: dict[str, dict[str, Any]]) -> list[dcc.Graph]:
    """Create plots for each subject by day taught.

    Args:
        stats_dict (dict[str, Any]): Dictionary of student data split by days taught.

    Returns:
        plots (list[Graph]): list of plots.

    """
    plots = []

    for subject, day_dicts in stats_dict.items():
        if subject in ["total_missed_time", "total_student_hours", "total_missed_days"]:
            continue

        data = []
        for day, info_dict in day_dicts.items():
            absent_pct = info_dict["total_missed_time"] / info_dict["total_student_hours"]
            absent_pct = round(100 * (1 - absent_pct), 2)

            data.append(
                {
                    "x": ["Attendance"],
                    "y": [absent_pct],
                    "type": "bar",
                    "name": day,
                }
            )

        graph = dcc.Graph(
            figure={
                "data": data,
                "layout": {"title": subject, "yaxis": {"range": [0, 100]}},
            },
        )
        plots.append(graph)

    return plots



# Import necessary libraries
import os
import warnings
from datetime import datetime
from typing import Any

import dash
import dash_bootstrap_components as dbc
import pandas as pd
import plotly.express as px
from dash import Output, callback, dcc, Input, html, dash_table

from config import settings
from pages.components.table import table_conditionals
from resources.dash_logic import calculate_data_table
from resources.dataframe import (
    process_dataframes,
    group_students_by_day,
    xlsx_student_data_to_dataframe,
)

warnings.simplefilter(action="ignore", category=FutureWarning)

dash.register_page(__name__, order=1)

xl_file = pd.ExcelFile(os.path.join(settings.DATA_DIR, settings.FILE_NAME))

dfs = {sheet_name: xl_file.parse(sheet_name) for sheet_name in xl_file.sheet_names}

dfs = process_dataframes(dataframe_dict=dfs)

df = pd.DataFrame(
    {
        "name": ["one", "two", "three"],
        "total_lessons": [3, 3, 3],
        "attended_lessons": [1, 2, 3],
        "missed_lessons": [1, 2, 3],
        "total_minutes_missed": [1, 2, 3],
        "attendance_percentage": [1, 2, 3],
        "subject": ["Monday", "Monday", "Wednesday"],
    }
)

SUBJECTS = list(dfs.keys())

# Define the page layout
layout = dbc.Container(
    [
        html.Br(),
        dbc.Row(
            [
                html.Center(html.H1("", id="title")),
            ],
            style={"margin-bottom": 25},
        ),
        # Row of dropdown menus
        dbc.Row(
            [
                dbc.Col(
                    [
                        dbc.DropdownMenu(
                            children=[dbc.DropdownMenuItem(sub, id=sub) for sub in SUBJECTS],
                            label="Subject",
                            color="primary",
                            className="m-1",
                            id="subject_dropdown",
                        ),
                    ],
                    style={"display": "flex", "flexWrap": "wrap"},
                ),
                dbc.Col([]),
                dbc.Col(
                    [
                        dcc.DatePickerRange(
                            start_date=datetime(2023, 1, 1),
                            end_date=datetime(2099, 1, 1),
                            start_date_id="start_date",
                            end_date_id="end_date",
                            id="date_picker",
                            display_format="DD-MM-YYYY",
                        ),
                    ]
                ),
            ]
        ),
        # Line plots of student attendance
        dbc.Row(
            dcc.Graph(id="line-chart", figure={}, className="twelve columns"),
            id="plots",
        ),
        # Table of student statistics
        dbc.Row(
            html.Div(
                dash_table.DataTable(
                    data=df.to_dict("records"),
                    sort_action="native",
                    columns=[{"name": i, "id": i} for i in df.columns],
                    style_data_conditional=table_conditionals(),
                    style_as_list_view=True,
                    style_header={
                        "backgroundColor": "rgb(30, 30, 30)",
                        "color": "white",
                    },
                    style_cell={"textAlign": "center"},
                    id="data_table_plot",
                ),
                className="row",
                style={"margin-bottom": 50},
            )
        ),
    ]
)


@callback(
    [
        Output(component_id="title", component_property="children"),
        Output(component_id="plots", component_property="children"),
        Output("data_table_plot", "data"),
    ],
    Input("date_picker", "start_date"),
    Input("date_picker", "end_date"),
    [Input(component_id=sub, component_property="n_clicks") for sub in SUBJECTS],
)
def pick_data_by_subject(
    start_date: str, end_date: str, *args: Any
) -> tuple[str, dbc.Row, list[dict[str, str | int | float]]]:
    """Output subject related data.

    Args:
        start_date (str): Earliest date filter.
        end_date (str): Latest date filter.
        args (tuple[Any]): Additional arguments from the dash context.

    Returns:
        data (tuple[subject, dbc.Row(plots), new_table_data]): Chosen subject name, plots of
        subject data, and table data for chosen subject.

    """
    ctx = dash.callback_context

    subject = ctx.triggered[0]["prop_id"].split(".")[0]

    if not subject or subject in ["date_picker", "start_date", "end_date"]:
        subject = SUBJECTS[0]

    # Get all student data for subject
    df = xlsx_student_data_to_dataframe(subject=subject)

    # Apply date range filter
    df = df[(df["date"] > start_date) & (df["date"] < end_date)]

    # Separate students according to days taught
    df_by_days = group_students_by_day(df=df)

    figures = {}

    for day, student_df in df_by_days.items():
        student_names = [
            col
            for col in student_df.columns
            if col not in ["date", "lesson time (minutes)", "day"]
        ]

        student_data_df = df[["date"] + student_names].copy()
        student_data_df["date"] = [dt.date() for dt in student_data_df["date"]]
        student_data_df.set_index("date", inplace=True)
        fig = px.scatter(student_data_df, x=student_data_df.index, y=student_names)
        fig.update_layout(yaxis_range=[0, max(df["lesson time (minutes)"]) * 1.1])
        figures[day] = fig

    plots = []

    for day, fig in figures.items():
        fig.update_layout({"title": day, "legend_title_text": "Student Name"})
        plots.append(dcc.Graph(id="line-chart", figure=fig, className="twelve columns"))
    new_table_data = calculate_data_table(dfs)
    new_table_data = [row for row in new_table_data if row["subject"] == subject]
    return subject, dbc.Row(plots), new_table_data



from config import settings


condition_dict = dict[str, str | dict[str, str | list[str]]]


def table_conditionals() -> list[condition_dict]:
    """Conditional statements for dash table.

    Returns
        conditionals (list[dict]): Conditional statements for dash table.
    """
    return [
        {
            "if": {
                "filter_query": f"{{attendance_percentage}} >= {settings.ATTENDANCE_GREEN}",
                "column_id": ["name", "attendance_percentage"],
            },
            "color": settings.COLOUR_GREEN_RGB,
        },
        {
            "if": {
                "filter_query": f"{{attendance_percentage}} >= {settings.ATTENDANCE_AMBER} && "
                f"{{attendance_percentage}} < {settings.ATTENDANCE_GREEN}",
                "column_id": ["name", "attendance_percentage"],
            },
            "color": settings.COLOUR_AMBER_RGB,
        },
        {
            "if": {
                "filter_query": f"{{attendance_percentage}} < {settings.ATTENDANCE_AMBER}",
                "column_id": ["name", "attendance_percentage"],
            },
            "color": settings.COLOUR_RED_RGB,
        },
    ]



import dash
from dash import html
import dash_bootstrap_components as dbc


# Define the navbar structure
def simple_navbar() -> html.Div:
    """Return the nav bar for the dash application."""
    layout = html.Div(
        [
            dbc.NavbarSimple(
                children=[
                    dbc.NavItem(dbc.NavLink(page["name"], href=page["relative_path"]))
                    for page in dash.page_registry.values()
                ],
                brand="Nicola Hearne",
                brand_href="/attendance",
                color="dark",
                dark=True,
            ),
        ]
    )

    return layout



# Import necessary libraries
import dash
import dash_bootstrap_components as dbc
from dash import html, dcc
from dash.dependencies import Input, Output


# Define the Dash App and it's properties here
app = dash.Dash(
    __name__,
    external_stylesheets=[dbc.themes.BOOTSTRAP],
    meta_tags=[{"name": "viewport", "content": "width=device-width"}],
    suppress_callback_exceptions=True,
)

from pages import about, subjects, overview
from pages.components import navbar

# define the navbar
nav = navbar.simple_navbar()

# Define the index page layout
app.layout = html.Div(
    [
        dcc.Location(id="url", refresh=False),
        nav,
        html.Div(id="page-content", children=[]),
    ]
)


@app.callback(Output("page-content", "children"), [Input("url", "pathname")])
def display_page(pathname: str) -> dbc.Container | str:
    """Display the specified page for the corresponding pathname.

    Args:
        pathname (str): Name of the page to display.

    Returns:
        result (dbc.Container | str): Container of the requested page, or a string message
         indicating a 404 error.

    """
    path_dict = {
        "/": subjects.layout,
        "/about": about.layout,
        "/subjects": subjects.layout,
        "/overview": overview.layout,
    }

    if layout := path_dict.get(pathname):
        return layout

    return "404 Page Error! Please choose a link"


# Run the app on localhost:8050
if __name__ == "__main__":
    app.run_server(debug=True)



import warnings

import dash
from dash import html
import dash_bootstrap_components as dbc

from resources.dash_logic import generate_subject_stats_dict, create_plots

warnings.simplefilter(action="ignore", category=FutureWarning)

dash.register_page(__name__, order=2)

# Generate necessary page statistics
subject_stats, days_taught, num_students, num_subjects = generate_subject_stats_dict()

layout = dbc.Container(
    [
        html.Br(),
        # Title
        dbc.Row(
            [
                html.Center(html.H1("Teaching Stats", id="student_title")),
            ],
            style={"margin-bottom": 25},
        ),
        # Teaching Stats
        dbc.Row(
            [
                dbc.Col(
                    [
                        html.Center(html.H3("Total Subjects")),
                        html.Center(html.H1(f"{num_subjects}")),
                    ],
                    style={"display": "row  ", "flexWrap": "wrap"},
                ),
                dbc.Col(
                    [
                        html.Center(html.H3("Days Teaching")),
                        html.Center(html.H1(f"{days_taught}")),
                    ],
                    style={"display": "row  ", "flexWrap": "wrap"},
                ),
                dbc.Col(
                    [
                        html.Center(html.H3("Total Students")),
                        html.Center(html.H1(f"{num_students}")),
                    ],
                    style={"display": "row  ", "flexWrap": "wrap"},
                ),
            ]
        ),
        # Line plots of student attendance by subject
        dbc.Row(create_plots(stats_dict=subject_stats)),
    ]
)



import dash_bootstrap_components as dbc
from dash import html


subjects_card = dbc.Card(
    [
        dbc.CardBody([html.Center(html.H4("Track Students By Subject", className="card-title"))]),
        dbc.CardImg(src="assets/static/images/subjects.png", bottom=True),
    ],
    style={"width": "70rem", "margin-bottom": 25},
)

table_card = dbc.Card(
    [
        dbc.CardBody([html.Center(html.H4("View/Organise Statistics", className="card-title"))]),
        dbc.CardImg(src="assets/static/images/table_stats.png", top=False),
    ],
    style={"width": "70rem", "margin-bottom": 25},
)

overview_card = dbc.Card(
    [
        dbc.CardBody([html.Center(html.H4("Visualise Teaching Load  ", className="card-title"))]),
        dbc.CardImg(src="assets/static/images/overview.png", top=False),
    ],
    style={"width": "70rem", "margin-bottom": 25},
)



import os

import pandas as pd

from config import settings


def add_day_of_week_column(df: pd.DataFrame) -> pd.DataFrame:
    """Add corresponding week day column to DataFrame.

    Expects existing column 'date' with datetime format YYYY-mm-dd.

    Args:
        df (pd.DataFrame): DataFrame of subject.

    Returns:
        df (pd.DataFrame): Original dataframe with new 'day' column.

    """
    dates = pd.to_datetime(df["date"], format="%d-%m-%Y")

    df["day"] = [dt.strftime("%A") for dt in dates]

    return df


def process_dataframes(dataframe_dict: dict[str, pd.DataFrame]) -> dict[str, pd.DataFrame]:
    """Preprocessing of dataframes.

    Steps include:
        1. Converting all column names to lower-case.
        2. Adding a 'day' column.

    Args:
        dataframe_dict (dict[str, pd.DataFrame]): Student info by subject separated into days.

    Returns:
        data_dict (dict[str, pd.DataFrame]): Processed student info.

    """
    data_dict = {}

    for sheet_name, df in dataframe_dict.items():
        df.columns = df.columns.str.lower()
        data_dict[sheet_name] = add_day_of_week_column(df=df)

    return data_dict


def group_students_by_day(df: pd.DataFrame) -> dict[str, pd.DataFrame]:
    """Group students into separate DataFrames according to the day they are taught.

    Args:
        df (pd.DataFrame): DataFrame of students by subject taught.

    Returns:
        df_by_days (dict[str, pd.DataFrame]): Dict of day, dataframe pairs for each day taught.
    """
    df_by_days = {}

    for day in df["day"].unique():
        # Group students by day
        temp_df = df[df["day"] == day].copy()

        # Remove students never in on those days
        temp_df.dropna(axis=1, how="all", inplace=True)

        df_by_days[day] = temp_df

    return df_by_days


def xlsx_student_data_to_dataframe(subject: str) -> pd.DataFrame:
    """Return DataFrame of student data given a subject.

    Args:
        subject (str): Lesson subject.

    Returns:
        dataframe (pd.DataFrame): DataFrame of student info relating to subject specified.

    """
    xl_file = pd.ExcelFile(os.path.join(settings.DATA_DIR, settings.FILE_NAME))

    dfs = {sheet_name: xl_file.parse(sheet_name) for sheet_name in xl_file.sheet_names}

    dfs = process_dataframes(dataframe_dict=dfs)

    return dfs[subject]



import os

from dotenv import load_dotenv
from pydantic_settings import BaseSettings

load_dotenv(dotenv_path=os.path.join(os.getcwd(), ".config"))


class BaseConfig(BaseSettings):
    """Object to hold project level configuration."""

    BASE_DIR: str = os.path.dirname(os.path.realpath(__file__))
    DATA_DIR: str = os.path.join(os.getcwd(), "data")
    FILE_NAME: str = "attendance.xlsx"

    # dashboard config

    # Attendance range colours
    ATTENDANCE_GREEN: float = 80
    ATTENDANCE_AMBER: float = 70

    # Attendance range to highlight
    COLOUR_GREEN_RGB: str = "rgb(0, 153, 51)"
    COLOUR_AMBER_RGB: str = "rgb(255, 102, 0)"
    COLOUR_RED_RGB: str = "rgb(255, 0, 0)"


settings = BaseConfig()



# Import necessary libraries
import dash
from dash import html, dcc
import dash_bootstrap_components as dbc

from pages.components.card import overview_card, subjects_card, table_card

dash.register_page(__name__, order=3)

layout = dbc.Container(
    [
        dbc.Row(
            [
                html.Center(html.H1("About")),
                html.Br(),
                html.Hr(),
                dbc.Col(
                    [
                        dcc.Markdown(
                            """
                            
                            Student tracker was designed to provide a visual and quantitative 
                            assessment of all students taught. By monitoring student performance 
                            over time it is possible to:
                            
                            * Spot potential issues early.
                            * Notice patterns (good and bad) that may exist.
                            * Gain a quantitative assessment of student performance.
                            * Be better equipped to remove any barriers to learning.
                            
                            Currently you can:
                            * View attendance for all students filtered by subject and date range.
                            * Organise student statistics by column values.
                            * Get an overview of teaching workload broken down by subject and 
                            days taught.
                            """
                        ),
                    ],
                    style={"margin-bottom": 25},
                ),
            ]
        ),
        dbc.Row(
            [
                dbc.Col(subjects_card, width="auto"),
            ],
            justify="center",
        ),
        dbc.Row(
            [
                dbc.Col(table_card, width="auto"),
            ],
            justify="center",
        ),
        dbc.Row(
            [
                dbc.Col(overview_card, width="auto"),
            ],
            justify="center",
        ),
    ]
)



import os
import shutil
import sys

from config import settings


def start_dash() -> None:
    """CLI to start dash application."""
    try:
        user_input = sys.argv[1]
    except IndexError:
        print("To start the dashboard type 'nicola dashboard'")
        sys.exit(1)

    if user_input == "init":
        print("*" * 50)
        print("*" * 14, "Initialising project", "*" * 14)
        print("*" * 50)
        cwd = os.getcwd()

        if not os.path.isdir(os.path.join(cwd, "data")):
            os.mkdir(os.path.join(cwd, "data"))

        print("\n\t* Created data directory")
        shutil.copy(
            src=os.path.join(os.path.join(settings.BASE_DIR, "data", "attendance.xlsx")),
            dst=os.path.join(cwd, "data", "attendance.xlsx")
        )
        print("\t* Copied example .xlsx file")
        shutil.copy(
            src=os.path.join(settings.BASE_DIR, "README.md"),
            dst=os.path.join(cwd, "README.md")
        )
        print("\t* Copied README.md")
        shutil.copy(
            src=os.path.join(settings.BASE_DIR, ".config"),
            dst=os.path.join(cwd, ".config")
        )
        print("\t* Copied .config file\n")

        print("*" * 50)
        print("*" * 13, "Project Setup Complete", "*" * 13)
        print("*" * 50)
        print("\n-  To start the dashboard type 'nicola dashboard'")
        print("-  Please Read 'README.md' for full instructions.\n")
        sys.exit(1)

    if user_input != "dashboard":
        print("To start the dashboard type 'nicola dashboard'")
        sys.exit(1)

    os.system(f"python {os.path.join(settings.BASE_DIR, 'app.py')}")
